import os
import subprocess
import traceback
from time import sleep

from google.cloud import storage
from google.cloud.storage import Blob
from googleapiclient.discovery import build

from gostep.consts import FUNCTIONS_API, AUTH_FILE
from gostep.consts import FUNCTIONS_API_VERSION
from gostep.consts import GCLOUD_STORAGE_CLASS


def get_service_client(api, version):
    """
        Creates and returns a service resource object.

            Parameters:
                api (string): api type of gcloud service
                version (string): gcloud service version

            Returns:
                service_client (object): A Resource object with methods for
                interacting with the service
    """
    try:
        return build(api, version)
    except Exception:
        print(traceback.format_exc())


def get_storage_client():
    """
        Creates and returns a storage service resource object.

            Returns:
                storage_client (object): storage service object
    """
    try:
        return storage.Client()
    except Exception:
        print(traceback.format_exc())


def get_locations(project):
    """
        Returns a list of available locations for a project.

            Parameters:
                project (string): root directory of the repo
                as 'projects/{project_id}'

            Returns:
                locations_list (list): list of location objects
    """
    try:
        service_client = get_service_client(
            FUNCTIONS_API, FUNCTIONS_API_VERSION)
        project_name = ''.join(['projects', '/', project])
        return service_client.projects().locations().list(name=project_name).execute()['locations']
    except Exception:
        print(traceback.format_exc())


def create_bucket(name, location):
    """
        Creates a storage bucket and returns the created bucket object.

            Parameters:
                name (string): name of the bucket
                location (string): location id

            Returns:
                bucket (object): instance of created bucket object
    """
    try:
        client = get_storage_client()
        bucket = storage.Bucket(name=name, client=client)
        bucket.location = location
        bucket.storage_class = GCLOUD_STORAGE_CLASS
        return client.create_bucket(bucket)
    except Exception:
        print(traceback.format_exc())


def get_buckets():
    """
        Returns a list of bucket objects.

            Returns:
                buckets (list): list of bucket objects
    """
    try:
        client = get_storage_client()
        return client.list_buckets()
    except Exception:
        print(traceback.format_exc())


def get_bucket(name):
    """
        Returns the bucket object.

            Parameters:
                name (string): name of the bucket

            Returns:
                bucket (object): instance of a bucket object
    """
    try:
        client = get_storage_client()
        return client.get_bucket(name)
    except Exception:
        print(traceback.format_exc())


def upload_file_to_bucket(bucket_name, file_name, file):
    """
        Uploads a file to a storage bucket and returns it's access path.

            Parameters:
                bucket_name (string): name of the bucket
                file_name (string): name of the file with extension
                file (string): path of the file to be uploaded

            Returns:
                file_path (string): bucket path of the uploaded file
    """
    try:
        bucket = get_bucket(bucket_name)
        blob = Blob(name=file_name, bucket=bucket)
        with open(file, 'rb') as file_object:
            blob.upload_from_file(file_object)
        return ''.join(['gs://', blob.bucket.name, '/', blob.name])
    except Exception:
        print(traceback.format_exc())


def get_cloud_functions(location_path):
    """
        Returns a object list which describes a cloud function in a project.

            Parameters:
                location_path (string): path to function including function id
                as 'projects/{project_id}/locations/{location_id}'

            Returns:
                functions_list (list): list of function objects
    """
    try:
        service_client = get_service_client(
            FUNCTIONS_API, FUNCTIONS_API_VERSION)
        functions = service_client.projects().locations().functions().list(
            parent=location_path
        ).execute()
        if functions == {} or functions == []:
            return []
        return functions['functions']
    except Exception:
        print(traceback.format_exc())


def get_cloud_function(function_path):
    """
        Returns a object which describes a cloud function in a project.

            Parameters:
                function_path (string): path to function including function id
                as projects/{project_id}/locations/{location_id}/functions/{id}


            Returns:
                function (object): a function object
    """
    try:
        service_client = get_service_client(
            FUNCTIONS_API, FUNCTIONS_API_VERSION)
        return service_client.projects().locations().functions().get(
            name=function_path
        ).execute()
    except Exception:
        print(traceback.format_exc())


def deploy_cloud_function(location_path, function_spec):
    """
        Deploys a function for a given specification and returns a function
        object.

            Parameters:
                location_path (string): path to function including function id
                as projects/{project_id}/locations/{location_id}
                function_spec (object): changes to be patched

            Returns:
                function (object): deployed function object
    """
    try:
        service_client = get_service_client(
            FUNCTIONS_API, FUNCTIONS_API_VERSION)
        return service_client.projects().locations().functions().create(
            location=location_path, body=function_spec).execute()
    except Exception:
        print(traceback.format_exc())


def update_cloud_function(function_path, patch_field, update_spec):
    """
        Creates a directory if it does not exist.

            Parameters:
                function_path (string): path to function including function id
                as projects/{project_id}/locations/{location_id}/functions/{id}
                patch_field (string): field to be patched
                update_spec (object): changes to be patched

            Returns:
                function (object): updated function object
    """
    try:
        service_client = get_service_client(
            FUNCTIONS_API, FUNCTIONS_API_VERSION)
        return service_client.projects().locations().functions().patch(
            name=function_path, updateMask=patch_field,
            body=update_spec
        ).execute()
    except Exception:
        print(traceback.format_exc())


def get_projects():
    """
        List down available projects.

            Returns:
                projects_list (object): list of project objects
    """
    try:
        response = str(subprocess.check_output('gcloud projects list', shell=True)).replace('\'', '').split('\\n')
        project_list = []
        for row in response:
            if response.index(row) == 0:
                continue
            project_list.append(row.split(' ')[0])
        return project_list
    except Exception:
        print(traceback.format_exc())


def get_iam_policy(resource):
    """
        Get IAM policy for a given resource.

            Parameters:
                resource: resource path as
                /projects/{}/locations/{}/functions/{}

            Returns:
                iam_policy (object): policy object
    """
    try:
        service_client = get_service_client(FUNCTIONS_API,
                                            FUNCTIONS_API_VERSION)
        return service_client.projects().locations().functions().getIamPolicy(
            resource=resource).execute()
    except Exception:
        print(traceback.format_exc())


def set_iam_policy(resource, policy_request):
    """
        Get IAM policy for a given resource.

            Parameters:
                resource (string): resource path as
                /projects/{}/locations/{}/functions/{}
                policy_request (policy object): policy dictionary object

            Returns:
                iam_policy (object): policy object
    """
    try:
        service_client = get_service_client(FUNCTIONS_API,
                                            FUNCTIONS_API_VERSION)
        return service_client.projects().locations().functions().setIamPolicy(
            resource=resource, body=policy_request).execute()
    except Exception:
        print(traceback.format_exc())


def get_service_account_email(account_name):
    """
        Returns service account name using account name.

            Parameters:
                account_name (string): service account name

            Returns:
                service_account (list): list containing a service account email
    """
    try:
        cmd = 'gcloud iam service-accounts list'
        accounts_list_proc = subprocess.check_output(cmd, shell=True)
        accounts_split = str(accounts_list_proc).replace('\\n', ' ').split()
        return [i for i in accounts_split if account_name in i and '.com' in i]
    except Exception:
        print(traceback.format_exc())


def create_credentials(name, project, display_name, workspace_dir):
    """
        Creates a credentials.json file using gcloud cli.

            Parameters:
                name (string): service account name
                project (string): gcloud project id
                display_name (string): service account display name
                workspace_dir (string): workspace directory path of the project

            Returns:
                succeed (boolean): status of account creation
    """
    try:
        account_email = get_service_account_email(name)
        if len(account_email) == 0:
            cmd = ''.join([
                'gcloud iam service-accounts create ', name, ' --display-name "',
                display_name, '"'])
            subprocess.check_output(cmd, shell=True)
            sleep(2)
            account_email = get_service_account_email(name)
        cmd = ''.join([
            'gcloud iam service-accounts keys create ', workspace_dir, '/',
            AUTH_FILE, ' --iam-account ', account_email[0]])
        subprocess.check_output(cmd, shell=True)
        cmd = ''.join(['gcloud projects add-iam-policy-binding ', project,
                       ' --member serviceAccount:', account_email[0],
                       ' --role "roles/owner"'])
        policy_bind_proc = subprocess.check_output(cmd, shell=True)
        return True if 'Updated' in str(policy_bind_proc) else False
    except Exception:
        print(traceback.format_exc())


def default_gcloud_project():
    """
        Call gcloud cli and returns default default selected project id.

            Returns:
                project_id (string): active gcloud project id
    """
    try:
        response = subprocess.check_output('gcloud config get-value project', shell=True)
        splits = str(response).replace('\\n', '').split("'")
        return splits[1] if len(splits) > 1 else ''
    except Exception:
        print(traceback.format_exc())


def set_credential_file(cred_file_path):
    """
        Set system env variable for GOOGLE_APPLICATION_CREDENTIALS.

            Parameters:
                cred_file_path (string): credentials file path
    """
    try:
        os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = cred_file_path
    except Exception:
        print(traceback.format_exc())



from gostep.aggregator import bootstrap_base
from gostep.aggregator import bootstrap_service
from gostep.aggregator import deploy
from gostep.aggregator import deploy_all
from gostep.aggregator import get_cloud_functions
from gostep.file_manager import get_checksum
from gostep.file_manager import get_json_from_file
from gostep.file_manager import get_json_from_file
from gostep.file_manager import path_exists
from gostep.gcloud_ops import create_credentials
from gostep.gcloud_ops import default_gcloud_project
from gostep.gcloud_ops import get_locations
from gostep.gcloud_ops import get_projects
from gostep.gcloud_ops import set_credential_file



from gostep.consts import CMD_TREE, VALIDATION_MESSAGES, TYPE, TEXT, REQUIRED_FIELDS, BOOLEAN, CMD_BRANCHES, COMMANDS


def print_messages(msgs):
    for line in msgs:
        print(line)


def validate_fields(arg_schema_branch, args):
    if len(arg_schema_branch[REQUIRED_FIELDS]) != 0:
        if isinstance(arg_schema_branch[REQUIRED_FIELDS], dict):
            arg_schema_field_keys = arg_schema_branch[REQUIRED_FIELDS].keys()
            if not all(i in args for i in arg_schema_field_keys):
                print_messages(arg_schema_branch[VALIDATION_MESSAGES])
                return False
            for field_key in arg_schema_field_keys:
                if arg_schema_branch[REQUIRED_FIELDS][field_key][TYPE] == TEXT:
                    field_arg_index = args.index(field_key)
                    if field_arg_index == len(args) - 1 or args[field_arg_index + 1] in COMMANDS:
                        print_messages(arg_schema_branch[VALIDATION_MESSAGES])
                        return False
        if isinstance(arg_schema_branch[REQUIRED_FIELDS], list):
            arg_schema_field_keys = [list(key.keys())[0] for key in arg_schema_branch[REQUIRED_FIELDS]]
            if not any(i in args for i in arg_schema_field_keys):
                print_messages(arg_schema_branch[VALIDATION_MESSAGES])
                return False
            if all(i in args for i in arg_schema_field_keys):
                print_messages(arg_schema_branch[VALIDATION_MESSAGES])
                return False
            for arg in arg_schema_field_keys:
                index = args.index(arg)
                if args[index != len(args) - 1] and args[index + 1] not in COMMANDS:
                    print_messages(arg_schema_branch[VALIDATION_MESSAGES])
                    return False
            return True
    return True


def validated(args=None):
    if args is None or len(args) <= 1:
        print_messages(CMD_TREE[VALIDATION_MESSAGES])
        return False
    if len(args) != len(set(args)):
        print_messages(
            [
                'Error: Gostep key words are not allowed to be used as variable names.',
                'Gostep keys:',
                '  ' + ', '.join(COMMANDS)
            ]
        )
        return False
    for arg_index in range(0, len(args)):
        if args[arg_index] in CMD_BRANCHES:
            arg_schema = CMD_TREE[args[arg_index]]
            if arg_index == len(args) - 1 and arg_schema[TYPE] == BOOLEAN:
                return True
            for arg_schema_key in arg_schema.keys():
                if arg_schema_key == VALIDATION_MESSAGES:
                    continue
                elif arg_schema_key == TYPE:
                    index = args.index(args[arg_index])
                    if arg_schema[arg_schema_key] == TEXT:
                        if index == len(args) - 1 or args[index + 1] in COMMANDS:
                            print_messages(arg_schema[VALIDATION_MESSAGES])
                            return False
                    elif arg_schema[arg_schema_key] == BOOLEAN and args[index + 1] in COMMANDS and args[index + 1] in arg_schema.keys() and arg_schema[args[index + 1]][TYPE] == BOOLEAN:
                        return True
                elif arg_schema_key == REQUIRED_FIELDS:
                    if not validate_fields(arg_schema, args):
                        return False
                    return True
                else:
                    if not arg_schema_key in args:
                        print_messages(arg_schema[VALIDATION_MESSAGES])
                        return False
                    else:
                        arg_schema_branch = arg_schema[arg_schema_key]
                        if arg_schema_branch[TYPE] == TEXT:
                            branch_index = args.index(arg_schema_key)
                            if branch_index == len(args) - 1 or args[branch_index + 1] in CMD_BRANCHES:
                                print_messages(arg_schema_branch[VALIDATION_MESSAGES])
                                return False
                            validate_fields(arg_schema_branch, args)
                        elif arg_schema_branch[TYPE] == BOOLEAN:
                            continue
            return True
    print_messages(CMD_TREE[VALIDATION_MESSAGES])
    return False



import fileinput
import json
import os
import shutil
import subprocess
import traceback
from pathlib import Path
from zipfile import ZipFile

import yaml

from gostep.consts import GOSTEP_IGNORE_FILE


def get_all_files_dict(root_dir=os.getcwd()):
    """
        Get directory path and file name mapped structure as a dictionary

            Parameters:
                root_dir (string): root directory path

            Returns:
                dir_dict (dictionary): dictionary object containing filename
                against directory path
    """
    try:
        dir_dict = {}
        for root, dirs, files in os.walk(root_dir):
            for file in files:
                dir_dict[str(Path(os.path.abspath(os.path.join(root, file)))
                             .parent)] = file
        return dir_dict
    except Exception:
        print(traceback.format_exc())


def get_dir(dir_name, root_path=os.getcwd()):
    """
        Creates a directory if it does not exist.

            Parameters:
                root_path (string): root directory of the repo
                dir_name (string): sublevel directory of the repo

            Returns:
                dir_path (string): path of downloaded template
    """
    try:
        dir_path = ''.join([root_path, '/', dir_name])
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
            print('Created directory %s' % dir_path)
        return dir_path
    except Exception:
        print(traceback.format_exc())


def path_exists(path):
    """
         Returns whether the path file or directory path exists.

             Parameters:
                 path (string): path to file or directory

             Returns:
                 exists (boolean): file or dir status
     """
    return os.path.exists(path)


def copy_dir(source, destination):
    """
        Copy a directory tree and returns destination path.

            Parameters:
                source (string): source containing root directory path
                destination (string): target root directory path

            Returns:
                destination (string): copied destination path
    """
    try:
        shutil.copytree(
            source, destination, ignore=shutil.ignore_patterns('.svn'))
        return destination
    except Exception:
        print(traceback.format_exc())


def create_compressed_file(name, config_dir, sources_dir, target_dir):
    """
        Creates a compressed zip file removing given list of files to be
        ignored.

            Parameters:
                name (string): name of the compressed file
                config_dir (string): path where gostep config files exists
                sources_dir (string): source directory path
                target_dir (string): target directory path

            Returns:
                target_file_path (string): compressed file path
    """
    try:
        target_file_path = ''.join([target_dir, '/', name, '.zip'])
        compressed_file = ZipFile(target_file_path, "w")
        ignore_list = get_yaml_dict(GOSTEP_IGNORE_FILE, config_dir).split(' ')
        for dir_name, sub_dirs, files in os.walk(sources_dir):
            len_dir_path = len(dir_name)
            for filename in files:
                file_path = os.path.join(dir_name, filename)
                if any(substring in file_path for substring in ignore_list):
                    continue
                compressed_file.write(file_path, file_path[len_dir_path:])
        compressed_file.close()
        print("Successfully created compressed file %s" % target_file_path)
        return target_file_path
    except Exception:
        print(traceback.format_exc())


def write_to_file(file_name, dir_path, content):
    """
        Rewrite or create and write to file.

            Parameters:
                file_name (string): file name
                dir_path (string): root path which file contains
                content (string): content to be written

            Returns:
                file_path (string): copied destination path
    """
    file_path = ''.join([dir_path, '/', file_name])
    try:
        with open(file_path, 'w') as file_object:
            file_object.write(content)
        return file_path
    except Exception:
        print(traceback.format_exc())


def replace_string_in_file(file_name, dir_path, source_str, target_str):
    """
        Replace a string in a file.

            Parameters:
                file_name (string): file name
                dir_path (string): root path which file contains
                source_str (string): string to be replaced
                target_str (string): string which replace source string

            Returns:
                file_path (string): updated file path
    """
    file_path = ''.join([dir_path, '/', file_name])
    try:
        with fileinput.FileInput(file_path, inplace=True, backup='.bkp'
                                 ) as file_object:
            for line in file_object:
                line.replace(source_str, target_str)
        return file_path
    except Exception:
        print(traceback.format_exc())


def get_yaml_dict(yaml_file, source_dir):
    """
        Reads YAML file to a dictionary.

            Parameters:
                yaml_file (string): file name with the extension
                source_dir (string): target directory

            Returns:
                yaml_dict (dictionary): dictionary object
    """
    try:
        with open(''.join([source_dir, '/', yaml_file])) as yaml_source:
            return yaml.load(yaml_source, Loader=yaml.FullLoader)
    except Exception:
        print(traceback.format_exc())


def build_yaml_file(yaml_file, source_dir, yaml_dict):
    """
        Writes dictionary object to yaml file.

            Parameters:
                yaml_file (string): file name with the extension
                source_dir (string): target directory
                yaml_dict (dictionary): dictionary object to get written

            Returns:
                yaml_dict (dictionary): updated dictionary object
    """
    try:
        with open(''.join([source_dir, '/', yaml_file]), 'w') as yaml_source:
            yaml.dump(yaml_dict, yaml_source)
        return get_yaml_dict(yaml_file, source_dir)
    except Exception:
        print(traceback.format_exc())


def get_json_from_file(json_file):
    """
        Reads and returns a JSON file as an dictionary object.

            Parameters:
                json_file (string): path to json file

            Returns:
                json_object (dictionary): json dictionary
    """
    try:
        with open(json_file) as json_file_object:
            return json.load(json_file_object)
    except Exception:
        print(traceback.format_exc())


def rewrite_json_file(json_file, json_dict):
    """
        Reads and returns a JSON file as an dictionary object.

            Parameters:
                json_file (string): path to json file
                json_dict (dictionary): data dictionary object

            Returns:
                json_object (dictionary): updated json from file
    """
    try:
        with open(json_file, 'w') as json_file_object:
            json.dump(json_dict, json_file_object, indent=4)
        return get_json_from_file(json_file)
    except Exception:
        print(traceback.format_exc())


def get_checksum(dir_path):
    """
        Reads and returns a md5 checksum of a directory.

            Parameters:
                dir_path (string): path to directory

            Returns:
                checksum (string): checksum value of string
    """
    try:
        print(dir_path)
        cmd = ''.join(['tar c "', dir_path, '" | md5sum'])
        result = subprocess.check_output(cmd, shell=True)
        return result.decode().split()[0]
    except Exception:
        print(traceback.format_exc())



import re
from os.path import join

from gostep.builders import build_java_project
from gostep.consts import BASE_CONFIG_FILE, BUILD_DIR, GOSTEP_BUCKET, SERVICES, \
    TEMPLATES, NAME, DESCRIPTION, VERSION, SOURCE_DIRECTORY, SOURCE_ARCHIVE, \
    LOCATION_NAME, LOCATION_ID, PROJECT_ID, DEFAULT_LOCATION, \
    ENVIRONMENT, AUTH_FILE, SERVICE_CONFIG_FILE, CHECKSUM, TRIGGER, RUNTIME, JAVA_RUNTIME, \
    ALLOW_ALL
from gostep.consts import TEMPLATE_DIRECTORY
from gostep.file_manager import copy_dir, get_checksum
from gostep.file_manager import get_dir
from gostep.file_manager import get_json_from_file, create_compressed_file, \
    rewrite_json_file
from gostep.gcloud_ops import get_cloud_functions, get_buckets, \
    create_bucket, get_bucket, update_cloud_function, deploy_cloud_function, \
    upload_file_to_bucket, get_locations, set_iam_policy, get_iam_policy, \
    get_cloud_function
from gostep.repo_service import clone_template


def bootstrap_base(workspace_dir, project_name, description, default_location,
                   version):
    """
        Writes main project configuration.

            Parameters:
                workspace_dir (string): current workspace directory
                project_name (string): name of the project
                description (string): a little about project
                default_location (string): default location id
                version (string): version of the project

            Returns:
                project_spec (dictionary): generated project specification
    """
    print('Creating project specification...')
    credentials = get_json_from_file(''.join([workspace_dir, '/', AUTH_FILE]).replace('\\', ''))
    default_location = default_location if default_location is not None else \
        get_locations(credentials[PROJECT_ID])[0]['locationId']
    project_info = {
        NAME: project_name,
        PROJECT_ID: credentials[PROJECT_ID],
        DESCRIPTION: description,
        DEFAULT_LOCATION: default_location,
        VERSION: version,
        TEMPLATES: {},
        SERVICES: {}
    }
    project_spec = rewrite_json_file(
        ''.join([workspace_dir, '/', BASE_CONFIG_FILE]), project_info)
    get_dir(TEMPLATE_DIRECTORY, workspace_dir)
    print('Project base %s has been successfully generated.'
          % project_spec[NAME])
    return project_spec


def get_template(workspace_dir, environment):
    """
        Get template from template store, download if it does not exists.

        Parameters:
            workspace_dir (string): workspace directory
            environment (string): runtime

        Returns:
            template_dir (string): path to template
    """
    template_dir = get_dir(TEMPLATE_DIRECTORY, workspace_dir)
    return get_dir(environment, template_dir)


def get_template_from_store(environment, trigger, workspace_dir, project_spec):
    """
        Gets template from local store, download if it does not exists locally.

        Parameters:
            workspace_dir (string): workspace directory
            environment (string): runtime
            trigger (string): function invocation trigger
            project_spec (dictionary): base project config

        Returns:
            template_dir (string): path to template in local store
    """
    template_id = ''.join([environment, '/', trigger])
    template_path = ''.join([TEMPLATE_DIRECTORY, '/', template_id])
    if template_id not in project_spec.get(TEMPLATES).keys():
        template_path = get_dir(template_path, workspace_dir)
        source_template = clone_template(template_id, template_path)
        project_spec[TEMPLATES][template_id] = source_template.replace(''.join([workspace_dir, '/']), '')
        rewrite_json_file(''.join([workspace_dir, '/', BASE_CONFIG_FILE]), project_spec)
        return source_template
    return join(workspace_dir, project_spec[TEMPLATES][template_id])


def bootstrap_service(workspace_dir, name, description, environment, location, version, trigger, allow_all=False):
    """
        Gets template and builds directory for a service.

        Parameters:
            workspace_dir (string): workspace directory
            name (string): name of the service
            description (string): additional description
            environment (string): runtime
            location (string): gcloud location id
            version (string): version number,
            trigger (string): function invocation trigger,
            allow_all (boolen): allow invoke access to public

        Returns:
            service_spec (object): dictionary object containing service info
    """
    project_spec_file = ''.join([workspace_dir, '/', BASE_CONFIG_FILE])
    project_spec = get_json_from_file(project_spec_file)
    service_name = re.sub('[^A-Za-z0-9]+', '-', name).lower()
    if service_name in project_spec[SERVICES].keys():
        print('Service already exists. Please select a different name')
        return project_spec[SERVICES][service_name][SOURCE_DIRECTORY]
    print(''.join(['Fetching template from store for ', service_name]))
    template_dir = get_template_from_store(environment, trigger, workspace_dir, project_spec)
    sources_root = get_dir('src', workspace_dir)
    source_path = copy_dir(template_dir, ''.join([sources_root, '/', service_name]))
    location_name = ''.join(['projects/', project_spec[PROJECT_ID], '/locations/', location])
    function_name = ''.join([location_name, '/functions/', service_name])
    function_spec_file = ''.join([source_path, '/', SERVICE_CONFIG_FILE])
    function_spec = get_json_from_file(function_spec_file)
    function_spec[NAME] = function_name
    function_spec[DESCRIPTION] = description
    function_spec = rewrite_json_file(function_spec_file, function_spec)
    print('Function specification for %s has been updated.' % function_spec[NAME])
    project_spec = get_json_from_file(project_spec_file)
    project_spec[SERVICES][service_name] = {
        NAME: service_name,
        DESCRIPTION: description,
        SOURCE_DIRECTORY: source_path.replace(''.join([workspace_dir, '/']), ''),
        SOURCE_ARCHIVE: '',
        LOCATION_NAME: location_name,
        LOCATION_ID: location,
        VERSION: version,
        ENVIRONMENT: environment,
        TRIGGER: trigger,
        CHECKSUM: '',
        ALLOW_ALL: allow_all
    }
    project_spec = rewrite_json_file(project_spec_file, project_spec)
    print(''.join(['Function has been configured in ', project_spec[SERVICES][service_name][SOURCE_DIRECTORY]]))
    return project_spec[SERVICES][service_name]


def cloud_function_exists(name, location_path):
    """
        Find that function already has been deployed.

        Parameters:
            name (string): name of the service
            location_path (string): location path of the functions

        Returns:
            is_exists (boolean): true, if function already has been deployed
    """
    cloud_functions = get_cloud_functions(location_path)
    return any(name == function[NAME] for function in cloud_functions)


def bucket_exists(name):
    """
        Find that storage bucket already has been deployed.

        Parameters:
            name (string): name of the bucket

        Returns:
            is_exists (boolean): true, if bucket has been created
    """
    buckets = list(get_buckets())
    return any(name == bucket.name for bucket in buckets)


def get_storage_bucket(name, location):
    """
        Get the storage bucket, create one if it does not exists.

        Parameters:
            name (string): name of the bucket
            location(string): storage bucket location

        Returns:
            bucket (object): storage bucket object
    """
    if not bucket_exists(name):
        return create_bucket(name, location)
    return get_bucket(name)


def authorize_public_invoking(resource_path):
    """
        Add IAM policy to resource, for public invoking.

        Parameters:
            resource_path (string): path to function as
            /projects/{}/locations/{}/functions/{}

        Returns:
            policy (object): updated policy
    """
    return set_iam_policy(resource_path, {
        'policy': {
            'bindings': [
                {
                    'role': 'roles/cloudfunctions.invoker',
                    'members': ['allUsers']
                }
            ]
        }
    })


def invoke_role_exists(resource_path):
    """
        Find that policies have been set for the resource.

        Parameters:
            resource_path (string): path to function as
            /projects/{}/locations/{}/functions/{}

        Returns:
            is_exists (boolean): true, if policy bindings exists
    """
    policy = get_iam_policy(resource_path)
    return 'bindings' in policy


def upload_source_to_bucket(workspace_dir, name, service_dir, location, runtime):
    """
        Build compressed file and upload it into storage bucket.

        Parameters:
            workspace_dir (string): workspace directory path
            name (string): name of the function
            service_dir (string): path to service source
            location (string): region id
            runtime (string): runtime environment

        Returns:
            source_url (string): path to cloud function
    """
    build_dir = get_dir(BUILD_DIR, workspace_dir)
    service_root = service_dir
    if JAVA_RUNTIME in runtime:
        service_dir = build_java_project(service_dir)
    source_archive = create_compressed_file(name, service_root, service_dir, build_dir)
    storage_bucket = get_storage_bucket(GOSTEP_BUCKET, location)
    return upload_file_to_bucket(storage_bucket.name, name, source_archive)


def deploy(name, location, workspace_dir):
    """
        Deploy a cloud function service, redeploy if already has been deployed.

        Parameters:
            name (string): service name
            location (string): region id
            workspace_dir (string): workspace directory path

        Returns:
            function (object): cloud function object
    """
    project_spec_file = ''.join([workspace_dir, '/', BASE_CONFIG_FILE])
    project_spec = get_json_from_file(project_spec_file)
    service_name = re.sub('[^A-Za-z0-9]+', '-', name).lower()
    if service_name not in project_spec[SERVICES].keys():
        print('Service does not exists.')
        return False
    service_info = project_spec[SERVICES][service_name]
    service_dir = ''.join([workspace_dir, '/', service_info[SOURCE_DIRECTORY]])
    location = location if location is not None else get_locations(
        project_spec[PROJECT_ID])[0]['locationId'] if project_spec[DEFAULT_LOCATION] == '' else project_spec[
        DEFAULT_LOCATION]
    location_name = ''.join(['projects/', project_spec[PROJECT_ID], '/locations/', location])
    function_name = ''.join([location_name, '/functions/', service_name])
    function_spec_file = ''.join([service_dir, '/', 'function.json'])
    function_spec = get_json_from_file(function_spec_file)
    source_archive_url = upload_source_to_bucket(workspace_dir, service_name, service_dir, location,
                                                 function_spec[RUNTIME])
    function_spec[NAME] = function_name
    function_spec['sourceArchiveUrl'] = source_archive_url
    if cloud_function_exists(function_name, location_name):
        result = update_cloud_function(function_spec[NAME], 'sourceArchiveUrl', function_spec)
        function_spec[NAME] = result['metadata']['target']
    else:
        result = deploy_cloud_function(location_name, function_spec)
        function_spec[NAME] = result['metadata']['target']
    if project_spec[SERVICES][service_name][ALLOW_ALL] and not invoke_role_exists(function_spec[NAME]):
        authorize_public_invoking(function_spec[NAME])
    rewrite_json_file(function_spec_file, function_spec)
    project_spec[SERVICES][service_name][LOCATION_NAME] = location_name
    project_spec[SERVICES][service_name][LOCATION_ID] = location
    rewrite_json_file(project_spec_file, project_spec)
    service = get_cloud_function(function_spec[NAME])
    print(''.join(['New deployment has been triggered for ', service_name]))
    return service


def deploy_all(workspace_dir):
    project_spec_file = ''.join([workspace_dir, '/', BASE_CONFIG_FILE])
    project_spec = get_json_from_file(project_spec_file)
    for service_key in project_spec[SERVICES].keys():
        service = project_spec[SERVICES][service_key]
        service_dir = ''.join([workspace_dir, '/', service[SOURCE_DIRECTORY]])
        service_checksum = get_checksum(service_dir)
        if service_checksum != service[CHECKSUM]:
            print("Deploying service %s..." % service[NAME])
            deploy(service[NAME], service[LOCATION_ID], workspace_dir)
            service_checksum = get_checksum(service_dir)
            project_spec[SERVICES][service_key][CHECKSUM] = service_checksum
            project_spec = rewrite_json_file(project_spec_file, project_spec)
            print(''.join([service[NAME], ' service checksum updated ', service_checksum, '.']))
    print('All up to date.')
    return project_spec



from os import path

from setuptools import setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gostep',
    long_description=long_description,
    long_description_content_type='text/markdown',
    scripts=[
        './bin/gostep'
    ],
    version='0.1.0-beta5',
    description='Serverless Templates Provider for Google Cloud platform',
    url='https://github.com/codimite/gostep',
    author='Lahiru Pathirage',
    author_email='lahiru@codimite.com',
    license='MIT',
    packages=['gostep'],
    zip_safe=False,
    install_requires=[
        'svn',
        'google-cloud-storage',
        'google-api-python-client',
        'google-auth-httplib2',
        'google-auth-oauthlib',
        'oauth2client',
        'pyyaml'
    ]
)



GOSTEP_VERSION = 'v0.1.0beta'

GCLOUD_STORAGE_CLASS = 'STANDARD'
FUNCTIONS_API = 'cloudfunctions'
FUNCTIONS_API_VERSION = 'v1'
SERVICE_ENTRY_POINT = 'main'
ENVIRONMENTS = [
    'python',
    'nodejs',
    'java/plain',
    'java/spring'
]

TEMPLATE_REPO = 'https://github.com/codimite/gostep-templates/trunk'
BASE_CONFIG_FILE = 'config.json'
AUTH_FILE = 'credentials.json'
SERVICE_CONFIG_FILE = 'function.json'
TEMPLATE_DIRECTORY = 'templates'
BUILD_DIR = '../build'
GOSTEP_IGNORE_FILE = '.gostepignore'
GOSTEP_BUCKET = 'gostep'

SERVICES = 'services'
TEMPLATES = 'templates'
NAME = 'name'
ALLOW_ALL = 'allow_all'
RUNTIME = 'runtime'
JAVA_RUNTIME = 'java'
DESCRIPTION = 'description'
ENVIRONMENT = 'env'
VERSION = 'version'
SOURCE_DIRECTORY = 'source_dir'
SOURCE_ARCHIVE = 'source_archive'
LOCATION_NAME = 'location_name'
LOCATION_ID = 'location_id'
DEFAULT_LOCATION = 'default_location'
KIND = 'kind'
PROJECT_ID = 'project_id'
SERVICE_ACCOUNT_EMAIL = 'service_account_email'
CHECKSUM = 'checksum'
TRIGGER = 'trigger'
HTTP = 'http'
HTTPS_TRIGGER_KEY = 'httpsTrigger'
EVENT_TRIGGER_KEY = 'eventTrigger'
EVENT_TYPE = 'eventType'
EVENT_TYPE_PUBSUB = 'cloud.pubsub'
EVENT_TYPE_STORAGE = 'cloud.storage'
RESOURCE = 'resource'
VALIDATION_MESSAGES = 'msgs'
REQUIRED_FIELDS = 'required'
TYPE = 'type'
TEXT = 'text'
BOOLEAN = 'boolean'
TRIGGERS = ['http', 'pubsub', 'storage']

COMMANDS = [
    'auth',
    'init',
    'inside',
    'projects',
    'location',
    'locations',
    'display-name',
    'show',
    'base',
    'location',
    'env',
    'service',
    'deploy',
    'gcloud',
    'trigger',
    'allow-all',
    'version'
]

CMD_BRANCHES = [
    'auth',
    'base',
    'deploy',
    'gcloud',
    'service'
]

CMD_TREE = {
    'auth': {
        TYPE: BOOLEAN,
        'init': {
            TYPE: TEXT,
            VALIDATION_MESSAGES: [
                'Error: Invalid command.\nUsage:'
                '  gostep auth init <project name>',
                '    Optional args:\n'
                '        display-name <service account display name>\n'
                '        inside <workspace directory>'
            ]
        },
        'inside': {
            TYPE: TEXT,
            VALIDATION_MESSAGES: [
                'Error: Invalid command.\nUsage:',
                '  gostep auth inside <workspace directory>'
            ]
        },
        'show': {
            TYPE: BOOLEAN
        },
        VALIDATION_MESSAGES: [
            'Error: Invalid command.\nUsage:',
            '  gostep auth init <project name>',
            '    Optional args:\n'
            '        display-name <service account display name>\n'
            '        inside <workspace directory>'
            '  gostep auth show',
            '    Optional args:\n'
            '        inside <workspace directory>'
        ]
    },
    'base': {
        TYPE: BOOLEAN,
        'init': {
            TYPE: TEXT,
            VALIDATION_MESSAGES: [
                'Error: Invalid command.\nUsage:'
                '  gostep base init <project name>',
                '    Optional args:\n'
                '        explains <project info>\n'
                '        inside <workspace directory>\n'
                '        location <gcloud region id>\n'
                '        version <project version>'
            ]
        },
        'show': {
            TYPE: BOOLEAN
        },
        VALIDATION_MESSAGES: [
            'Error: Invalid command.\nUsage:',
            '  gostep base init <project name>',
            '    Optional args:\n'
            '        explains <project info>\n'
            '        inside <workspace directory>\n'
            '        location <gcloud region id>\n'
            '        version <project version>'
            '  gostep base show',
            '    Optional args:\n'
            '        inside <workspace directory>'
        ]
    },
    'deploy': {
        TYPE: TEXT,
        VALIDATION_MESSAGES: [
            'Error: Invalid command.\nUsage:',
            '  gostep deploy diff',
            '    Optional args:\n'
            '        inside <workspace directory>'
            '  gostep deploy <service name>',
            '    Optional args:\n'
            '        inside <workspace directory>'
        ]
    },
    'gcloud': {
        TYPE: BOOLEAN,
        REQUIRED_FIELDS: [
            {
                'projects': {
                    TYPE: BOOLEAN
                }
            },
            {
                'locations': {
                    TYPE: BOOLEAN
                }
            }
        ],
        VALIDATION_MESSAGES: [
            'Error: Invalid command.\nUsage:',
            '  gostep gcloud projects',
            '  gostep gcloud locations',
            '    Optional args:\n'
            '        inside <workspace directory>'
        ]
    },
    'service': {
        TYPE: BOOLEAN,
        'init': {
            TYPE: TEXT,
            REQUIRED_FIELDS: {
                'env': {
                    TYPE: TEXT,
                },
                'trigger': {
                    TYPE: TEXT
                }
            },
            VALIDATION_MESSAGES: [
                'Error: Invalid command.\nUsage:',
                '  gostep service init <service name> env <runtime environment> trigger <function invoker>',
                '    Optional args:\n'
                '        explains <project info>\n'
                '        inside <workspace directory>\n'
                '        location <gcloud region id>\n'
                '        version <project version>\n'
                '        allow-all'
            ]
        },
        VALIDATION_MESSAGES: [
            'Error: Invalid command.\nUsage:',
            '  gostep service init <service name> env <runtime environment> trigger <function invoking type>',
            '    Optional args:\n'
            '        explains <project info>\n'
            '        inside <workspace directory>\n'
            '        location <gcloud region id>\n'
            '        version <project version>\n'
            '        allow-all'
        ]
    },
    VALIDATION_MESSAGES: [
        'GOSTEP - Serverless templates provider for Google cloud platform',
        'Version: %s' % GOSTEP_VERSION,
        'Usage:',
        '  gostep auth init <gcloud service account name>',
        '  gostep auth inside <workspace directory>',
        '  gostep auth show',
        '  gostep base init <project name>',
        '  gostep base show',
        '  gostep deploy diff',
        '  gostep deploy <service name>',
        '  gostep gcloud locations',
        '  gostep gcloud projects',
        '  gostep service init <service name> env <runtime environment> trigger <function invoking type>'
    ]
}
